/* testedb.c -- this tests some of the empire database access functions */

#include <stdio.h>

#include "edbproc.h"
#include "edbpriv.h"
#include "misc.h"
#include "file.h"
#include "xy.h"
#include "nsc.h"
#include "genitem.h"
#include "edb.h"
#include "collect.h"

#define BUFSZ 1024


Empire_Database *edb = 0;
int debug = 0;

char *defaultInput = "csync.dat";
char *fname = 0;
	
int parseArgs(int argc, char **argp)
{
	int i;
	char c, *p;
	for (i=1;i<argc;i++) {
		p = argp[i];
		if (*p == '-') {
		    p++;
		    while ((c=*p++)!=0) {
		    	switch (c) {
		    	    case 'd':
		    	    	debug++;
		    	    	break;
		    	    default:
		    	    	perror("bad command line arg");
		    	}
		    }
		}
		else if (fname)
			perror("too many arguments");
		else
		    fname = p;
	}

}

void SectorClass (obj)
struct genobject *obj;
{
    struct dchrstr *dp = (struct dchrstr *)obj;
    printf ("Sector Class\n");
    printf ("\t%c %s\n", dp -> d_mnem, dp -> d_name);
}

void SectorObj (obj)
struct genobject *obj;
{
    struct sctstr *sp = (struct sctstr *)obj;
    printf ("Sector Object\n");
    printf ("\t%d,%d %d %d%%\n", sp -> sct_x, sp -> sct_y,
	    sp->sct_type, sp->sct_effic);
}

void geno (obj)
struct genobject *obj;
{
    printf("geno %d\n", obj -> ef_type);
}

void ShipClass (obj)
struct genobject *obj;
{
    struct mchrstr *mp = (struct mchrstr *)obj;
    printf ("Ship Class\n");
    printf ("\t%s cost %d tech %d\n", mp -> m_name, mp->m_cost, mp->m_tech);
}

void ShipObject (obj)
struct genobject *obj;
{
    struct shpstr *mp = (struct shpstr *)obj;
    printf ("Ship Object\n");
    printf ("\t%d,%d type %d eff %d%%\n", mp -> shp_x, mp -> shp_y,
	    mp -> shp_type, mp -> shp_effic);
}
void LandClass (obj)
struct genobject *obj;
{
    struct lchrstr *lp = (struct lchrstr *)obj;
    printf ("Land Class\n");
    printf ("\t%s cost %d tech %d\n", lp -> l_name, lp->l_cost, lp->l_tech);
}

void LandObject (obj)
struct genobject *obj;
{
    struct lndstr *lp = (struct lndstr *)obj;
    printf ("Land Object\n");
    printf ("\t%d,%d type %d eff %d%%\n", lp -> lnd_x, lp -> lnd_y,
	    lp -> lnd_type, lp -> lnd_effic);
}
void PlaneClass (obj)
struct genobject *obj;
{
    struct plchrstr *pp = (struct plchrstr *)obj;
    printf ("Plane Class\n");
    printf ("\t%s cost %d tech %d\n", pp -> pl_name, pp->pl_cost, pp->pl_tech);
}

void PlaneObject (obj)
struct genobject *obj;
{
    struct plnstr *pp = (struct plnstr *)obj;
    printf ("Plane Object\n");
    printf ("\t%d,%d type %d eff %d%%\n", pp -> pln_x, pp -> pln_y,
	    pp -> pln_type, pp -> pln_effic);
}
void NukeClass (obj)
struct genobject *obj;
{
    struct nchrstr *pp = (struct nchrstr *)obj;
    printf ("Nuke Class\n");
    printf ("\t%s cost %d tech %d\n", pp -> n_name, pp->n_cost, pp->n_tech);
}

void NukeObject (obj)
struct genobject *obj;
{
    struct nukstr *pp = (struct nukstr *)obj;
    printf ("Nuke Object\n");
    printf ("\t%d,%d\n", pp -> nuk_x, pp -> nuk_y);

}

void NatObject (obj)
struct genobject *obj;
{
    struct country *cp = (struct country *)obj;
    printf ("Nation Object\n");
    printf ("\tcountry %d %d$\n", cp -> fld_cnum, cp -> fld_cnam,
	    cp -> fld_money);
}
void VersionClass (obj)
struct genobject *obj;
{
    struct gameinfo *gp = (struct gameinfo *)obj;
    printf ("Version Class\n");
    printf ("\tMAXNOC %d WX %d WY %d\n",
	    gp -> fld_MAXNOC, gp -> fld_WORLD_X, gp -> fld_WORLD_Y);
}

void ItemClass (obj)
struct genobject *obj;
{
    struct ichrstr *ip = (struct ichrstr *)obj;
    printf ("Item Class\n");
    printf ("\tmnem %c %s\n", ip -> i_mnem, ip -> i_name);
}
void ProdClass (obj)
struct genobject *obj;
{
    struct pchrstr *pp = (struct pchrstr *)obj;
    printf ("Product Class\n");
    printf ("\t%s %s\n", pp -> p_name, pp -> p_sname);
}
void BMapObject (obj)
struct genobject *obj;
{
    struct bmapstr *bp = (struct bmapstr *)obj;
    printf ("BMap Object\n");
    printf ("\t%d,%d %s\n", bp -> fld_leftx, bp -> fld_y, bp -> fld_row);
}
void MapObject (obj)
struct genobject *obj;
{
    struct cmapstr *mp = (struct cmapstr *)obj;
    printf ("Map Object\n");
    printf ("\t%d,%d %s\n", mp -> fld_x, mp -> fld_y, mp -> fld_des);
}

struct disptable {
    int ef_type;
    void (*dfunc) (struct genobject *obj);
};

struct disptable DispTable[] = {
    { SCT_CLASS, SectorClass },
    { SCT_OBJECT, SectorObj },
    { LND_CLASS, LandClass },
    { LND_OBJECT, LandObject },
    { SHP_CLASS, ShipClass },
    { SHP_OBJECT, ShipObject },
    { PLN_CLASS, PlaneClass },
    { PLN_OBJECT, PlaneObject },
    { NUK_CLASS, NukeClass },
    { NUK_OBJECT, NukeObject },
    { ITM_CLASS, ItemClass },
    { PRD_CLASS, ProdClass },
    { VER_CLASS, VersionClass },
    { BMP_OBJECT, BMapObject },
    { MAP_OBJECT, MapObject },
    { NAT_OBJECT, NatObject },
    { -1, NULL},
};

void printobj (edb, obj)
Empire_Database *edb;
struct genobject *obj;
{
    struct disptable *dp;

    if (obj == NULL) {
	printf ("Empty object\n");
	return;
    }
    for (dp = DispTable; dp -> dfunc; dp ++)
	if (obj -> ef_type == dp -> ef_type) {
	    (*dp -> dfunc) (obj);
	    return;
	}

    dumpObject(edb, obj);
}

int main(int argc, char **argp)
{
	FILE *f;
	char buf[BUFSZ];
	struct DumpTable *p;
	struct genobject *obj;

	parseArgs(argc, argp);

	if (!fname) fname = defaultInput;	
	f = fopen(fname,"r+");
	if (!f) {
		perror("cannot open csync.dat");
		exit(1);
	}
	edb = edb_new();
	while (fgets(buf, BUFSZ, f)) {
	    obj = edb_parse_c_sync(edb, buf+2);
	    printobj(edb, obj);
	}
	fclose(f);
	printobj (edb, edb_getVersion (edb));
	printobj (edb, edb_getCountry (edb));
	exit(0);
}
