/*
	$Header: /a/victor/nexor/user4/jpo/xemp/xemp5.0/lib/ship/RCS/mission.c,v 1.9 1996/04/23 07:56:21 jpo Exp $
	$Date: 1996/04/23 07:56:21 $
	$Author: jpo $
	$Id: mission.c,v 1.9 1996/04/23 07:56:21 jpo Exp $
	$Locker:  $
	$Log: mission.c,v $
	Revision 1.9  1996/04/23 07:56:21  jpo
	New API

	Revision 1.8  1996/04/18 07:00:13  jpo
	Arbitrary sets of things to a mission

	Revision 1.7  1996/03/29 10:00:11  jpo
	Lint fixups

	Revision 1.6  1996/01/29 08:31:11  jpo
	New menu stuff

	Revision 1.5  1995/09/10 17:44:13  jpo
	Added prototype

	Revision 1.4  1995/09/10 16:22:46  jpo
	Changed missions to be more generic
	added support for fleet mode census and setting

	Revision 1.3  1995/09/08 07:43:14  jpo
	const

	Revision 1.2  1995/08/15 07:54:38  jpo
	Added lots of mission support

 * Revision 1.1  93/03/14  16:51:36  etienne
 * Initial revision
 * 
*/

#include "main.h"
#include "sector.h"
#include "ship.h"

static const char *mission_types[] = {
	"Delete Mission",
	"Interdict",
	(char *)0
};
#define MISSION_MAX	2

#define HOW_SELECT	0	/* via menu */
#define HOW_CURRENT	1	/* via redo */
#define HOW_BOTH	2	/* via bind */

static const char type2mission[] = "\0i";

static bool SetShipMission _PROTO((const char *ships, int range,
				   Sector sct, char *missp, int *rangep));

static bool SetShipMission (ships, range, sct, missp, rangep)
const char *ships;
int range;
Sector sct;
char *missp;
int *rangep;
{
	int i;
	Strings strings;
	char *answer;

	strings = InitStrings ();
	AddStringsID (strings, mission_types, MISSION_MAX, 0);
	i = ChooseMenu (strings, "Mission Type?", map_win, 5, 5, NULL);

	FreeStrings (strings);

	if (i < 0 || interrupt) {
		interrupt = False;
		Message ("Set mission cancelled");
		return False;
	}
	
	if (i == 0) {
		*missp = type2mission[i];
		*rangep = 0;
		FeedCommand (Fmt ("mission ship %s c", ships), PRINT);
		return True;
	}
	answer = InputAtMessage (Fmt  ("Range [%d]: ",range),
				 3, GS_PNUMBER);
	if (interrupt || answer == (char *)0) {
		interrupt = False;
		Message ("Mission cancelled");
		return False;
	}
	if (*answer != '\0')
		range = atoi (answer);
		
	*rangep = range;
	*missp = type2mission[i];
	FeedCommand (Fmt ("mission ship %s %c %s %d",
			  ships, *missp,
			  CrdStr(sct), range), PRINT);
	return True;
}

void SMission (list)
const char *list;
{
	int range = 0;
	int r;
	char miss;
	Ship ptr, aship;

	UnmarkAllShips (MARK_SELECT);
	MarkShips (list, MARK_SELECT);

	if (MarkedColocated (MARK_SELECT) == False) {
		Message ("Ships are not colocated");
		return;
	}

	aship = (Ship)0;
	for ALL_SHIPS(ptr) {
		if (sh_marked (ptr, MARK_SELECT)) {
			r = ShipFireRange (ptr);
			if (r > range)
				range = r;
			aship = ptr;
		}
	}

	if (aship == (Ship) 0) {
		Message ("No ships selected");
		return;
	}
	
	if (SetShipMission (list, range, ShipSector(aship),
			    &miss, &range) == False)
		return;
	for ALL_SHIPS(ptr) {
		if (sh_marked (ptr, MARK_SELECT)) {
			set_sh_mission (ptr, miss);
			ptr -> radius = range;
		}
	}
}

void FScanShipMission (list, flag)
const char *list;
int flag;
{
	FeedEmpire (Fmt ("mission ship %s q", list), flag);

	ScanShipMission (flag);
	(void) WaitForPrompt (flag);
}

void ScanShipMission (flag)
int flag;
{
	char *ptr, *ind;
	int mind, sno;
	Ship ship;

	if (EmpireStatus () != E_PRINTING)
		return;
	if (flag == PRINT)
		PrintAtEmpire ("Empiretool: scanning ship missions");

	ptr = ReadEmpire (DONT_PRINT);
	if ((ind = StrStr (ptr, "mission")) == NULL) {
		PrintAtEmpire (Fmt ("Mission scan failed on %s", ptr));
		return;
	}
	mind = ind - ptr;
	
	for (;;) {
		ptr = ReadEmpire (DONT_PRINT);

		if (EmpireStatus () != E_PRINTING) {
			if (flag == PRINT)
				PrintAtEmpire (ptr);
			return;
		}
		
		if ((ind = index (ptr, '#')) == NULL)
			continue;

		ind ++;
		if (!ScanDigit ((ConstVP)&ind, &sno) ||
		    (ship = NrToShip (sno)) == (Ship)0) {
			PrintAtEmpire (Fmt ("#%d Unknown ship (rescan ships!)",
					    sno));
			Bell ();
			(void) EmpireMore ();
			continue;
		}
		if (match_str (ptr, "has no mission", NULL, 0) == 0)
		    set_sh_mission (ship, '\0');
		else 
		    switch (ptr[mind]) {
		    case 'i':
			set_sh_mission (ship, 'i');
			break;
		    case 'h':
			set_sh_mission (ship, '\0');
			break;
		    default:
			PrintAtEmpire (Fmt ("Unknown mission type %s",
					    ptr + mind));
			Bell ();
			break;
		    }
	}
}

const char *FleetHasMission (fleet)
char fleet;
{
	Ship ptr;
	char mistype = '~';	/* bogus */
	const char *cp;

	for (ptr = shiplist; ptr != (Ship) 0; ptr = ptr-> next)
		if (ptr-> fleet == fleet && sh_mission(ptr) != mistype) {
			if (mistype == '~')
				mistype = sh_mission(ptr);
			else
				return "Mixed";
		}
	if (mistype == '\0')
		return "None";

	for (cp = type2mission + 1; *cp; cp++)
		if (*cp == mistype)
			return mission_types[cp - type2mission];
	return "Unknown";
}

void ShipMission (x, y)
int x, y;
{
    DoShipMission (x, y, HOW_SELECT);
}

void DoShipMission (x, y, how)
int x, y;
int how;
{
    	char * ptr;

	if (last_layout == SHIP_MODE && how != HOW_SELECT)
	{
		if (! sh_owned (curship))
		{
			Message ("Not your ship!");
			Bell ();
			return;
		}

		SMission (Fmt("%d", sh_nr(curship)));
	}
	else if (last_layout == FLEET_MODE && how != HOW_SELECT)
		SMission (Fmt("%c", curfleet));
	else
	{
		if (how == HOW_CURRENT)
		{
			Message ("No current ship/fleet");
			Bell ();
			return;
		}

		ptr = SelectShips (x, y, ST_ORDERS,
				   "Missions for which ships ?");
		if (ptr)
			SMission (ptr);
	}
}
