/*
	$Header: /a/victor/nexor/user4/jpo/xemp/xemp5.0/lib/ship/RCS/retreat.c,v 1.7 1996/04/18 07:00:38 jpo Exp $
	$Date: 1996/04/18 07:00:38 $
	$Author: jpo $
	$Id: retreat.c,v 1.7 1996/04/18 07:00:38 jpo Exp $
	$Locker:  $
	$Log: retreat.c,v $
	Revision 1.7  1996/04/18 07:00:38  jpo
	Arbitrary sets of things to a retreat

	Revision 1.6  1996/03/29 10:01:26  jpo
	Lint fixups

	Revision 1.5  1996/01/29 08:32:01  jpo
	New menu stuff
	plus some fixes

	Revision 1.4  1995/09/10 17:44:03  jpo
	added prototype and made conditions more generic

	Revision 1.3  1995/09/10 16:25:28  jpo
	Generic support
	added fleet mode settings

	Revision 1.2  1995/09/08 07:43:36  jpo
	retreat stuff

	Revision 1.1  1993/03/14 16:51:38  etienne
	Initial revision

*/

#include "main.h"
#include "sector.h"
#include "ship.h"
#include "func.h"

#define HOW_SELECT	0	/* via menu */
#define HOW_CURRENT	1	/* via redo */
#define HOW_BOTH	2	/* via bind */

const char *retreatcond[] = {
	"When damaged",
	"When torpedoed",
	"When bombed",
	"When sonar is detected",
	"When depth-charged",
	"When helpless",
	(char *)0
};


const char retreat_n2cond[] = "ITBSDH";

static bool ShipsRetreat _PROTO((const char *ships, uchar cond,
				 Sector sct, short *rtrxp, short *rtryp,
				 uchar *condp));

static bool ShipsRetreat (ships, cond, sct, rtrxp, rtryp, condp)
const char *ships;
uchar cond;
Sector sct;
short *rtrxp, *rtryp;
uchar *condp;
{
	Strings strings;
	char buf[10];
	char *cp;
	int nr;
	Sector to;
	int i;
	char *path;

	strings = InitStrings ();
	AddStringsID (strings, retreatcond,  RETREAT_MAXSTR, 0);
	MakeStringIndex (strings);
	for (i = 0; i < RETREAT_MAX; i++)
		if (cond & (1 << i))
			SetIthFlag (strings, i, 1);
	
	switch (MultiSelectPager (strings, "Retreat Conditions",
				  "Delete", DELETEFUNC, NULL)) {
	case DELETEFUNC:
		FreeStrings (strings);
		FeedCommand ( Fmt ("retreat %s h C", ships),
			      PRINT);
		*condp = RETREAT_NONE;
		*rtrxp = *rtryp = 0;
		return True;

	default:
	case CANCELFUNC:
		FreeStrings (strings);
		return False;

	case DONEFUNC:
		break;
	}
	InitStringList (strings);
	cp = buf;
	cond = RETREAT_NONE;
	while ((nr = GetNextSelectedId (strings)) != -1) {
		*cp ++ = retreat_n2cond[nr];
		cond |= (1 << nr);
	}
	*cp = '\0';
	FreeStrings (strings);

	Message (Fmt ("Retreat %s to sector?", ships));
	to = ChooseSeaDest (sct, &path);
	if (to == NULL) {
		return False;
	}
	*condp = cond;
	*rtrxp = s_xcd (to);
	*rtryp = s_ycd (to);
	FeedCommand ( Fmt ("retreat %s %s %s", ships, path, buf),
		     PRINT);
	return True;
}

void SRetreat (list)
const char *list;
{
	short rx, ry;
	uchar rc;
	int cond = 0;
	Ship aship = (Ship)0;
	Ship ptr;

	UnmarkAllShips (MARK_SELECT);
	MarkShips (list, MARK_SELECT);

	if (MarkedColocated(MARK_SELECT) == False) {
		Message ("Ships are not colocated");
		return;
	}

	for ALL_SHIPS(ptr) {
		if (sh_marked (ptr, MARK_SELECT)) {
			if (aship == (Ship)0) {
				aship = ptr;
				cond = aship -> rtrcond;
			}
			else if (cond != aship -> rtrcond)
				cond = 0;
		}
	}
	if (aship == (Ship)0) {
		Message ("No ships selected");
		return;
	}
	if (ShipsRetreat (list, cond,
			  World(sh_xcd(aship), sh_ycd(aship), S_DESIG),
			  &rx, &ry, &rc) == False)
		return;
	for ALL_SHIPS(ptr) {
		if (sh_marked (ptr, MARK_SELECT)) {
			ptr -> rtrcond = rc;
			ptr -> rtrtx = rx;
			ptr -> rtrty = ry;
		}
	}
}

void BindShipRetreat (x, y, str)
int x, y;
char *str;
{
	static char buf [100];

	if (sscanf (str, " ( %s ) ", buf) == 1) {
		SRetreat (buf);
	}
	else
		DoShipRetreatOrder (x, y, HOW_BOTH);
}

void ShipRetreat (x, y)
int x, y;
{
	DoShipRetreatOrder (x, y, HOW_SELECT);
}

void DoShipRetreatOrder (x, y, how)
int x, y;
int how;
{
    	char * ptr;

	if (last_layout == SHIP_MODE && how != HOW_SELECT)
	{
		if (! sh_owned (curship))
		{
			Message ("Not your ship!");
			Bell ();
			return;
		}

		SRetreat (Fmt("%d", sh_nr(curship)));
	}
	else if (last_layout == FLEET_MODE && how != HOW_SELECT)
		SRetreat (Fmt("%c", curfleet));
	else
	{
		if (how == HOW_CURRENT)
		{
			Message ("No current ship/fleet");
			Bell ();
			return;
		}

		ptr = SelectShips (x, y, ST_ORDERS, "Retreat which ships ?");
		if (ptr)
			SRetreat (ptr);
	}
}

void FScanShipRetreat (list, flag)
const char *list;
int flag;
{
	FeedEmpire (Fmt ("retreat %s", list), flag);

	ScanShipRetreat (flag);
	(void) WaitForPrompt (flag);
}

void ScanShipRetreat (flag)
int flag;
{
	char *ptr, *ind;
	Ship ship;
	int sno;
	int flagoff, pathoff;

	if (EmpireStatus () != E_PRINTING)
		return;
	if (flag == PRINT)
		PrintAtEmpire ("Empiretool: scanning ship retreats");
	
	ptr = ReadEmpire (DONT_PRINT);
	if ((ind = StrStr (ptr, "ship type")) == NULL) {
		PrintAtEmpire (Fmt ("Order scan failed on %s", ptr));
		return;
	}
	if ((ind = StrStr (ptr, "path")) == NULL) {
		PrintAtEmpire (Fmt ("No path in %s", ptr));
		return;
	}
	pathoff = ind - ptr;

	if ((ind = StrStr (ptr, "flags")) == NULL) {
		PrintAtEmpire (Fmt ("No flags in %s", ptr));
		return;
	}
	flagoff = ind - ptr;
	for (;;) {
		Sector to;
		ind = ptr = ReadEmpire (DONT_PRINT);

		if (EmpireStatus () != E_PRINTING) {
			if (flag == PRINT)
				PrintAtEmpire (ptr);
			return;
		}
		
		if (ScanDigit ((ConstVP)&ind, &sno)) /* ship name? */
			continue;
		if (strncmp (ind, "ship", 4) == 0) /* %d ships */
			break;
		if (strlen(ptr) < flagoff)
			continue;
		if ((ship = NrToShip (sno)) == (Ship)0) {
			PrintAtEmpire (Fmt ("#%d Unknown ship (rescan ships!)",
					    sno));
			Bell ();
			(void) EmpireMore ();
			continue;
		}
		ship -> rtrcond = RETREAT_NONE;
		for (ind = &ptr[flagoff]; *ind; ind ++) {
			char *cp;
			if ((cp = index (retreat_n2cond, *ind)) == NULL)
				continue;
			ship -> rtrcond |= (1 << (cp - retreat_n2cond));
		}
		ind = &ptr[pathoff];
		to = FollowPath (World(sh_xcd(ship), sh_ycd(ship), S_DESIG),
				 ind);
		ship -> rtrtx = s_xcd (to);
		ship -> rtrty = s_ycd (to);
	}
}

const char *RtrtList (cond)
uchar cond;
{
	static char buf[10], *cp;
	int i;

	for (cp = buf, i = 0; i < RETREAT_MAX; i++)
		if (cond & (1 << i))
			*cp++ = retreat_n2cond[i];
	*cp = '\0';
	return buf;
}

const char *FleetHasRetreat (fleet)
char fleet;
{
	Ship ptr;
	uchar rcond = 0;
	int first = 0;

	for (ptr = shiplist; ptr != (Ship) 0; ptr = ptr-> next)
		if (ptr-> fleet == fleet) {
			if (first++ == 0) 
				rcond = sh_retreat(ptr);
			else if (rcond != sh_retreat(ptr))
				return "Mixed";
		}
	
	if (rcond == RETREAT_NONE)
		return "No";
	return RtrtList (rcond);
}
